package pkg

import (
	"context"
	"errors"
	"github.com/julienschmidt/httprouter"
	"html/template"
	"net/http"
	"net/url"
	"strconv"
)

const (
	ContextCompanyKey = "numerus-company"
)

type Company struct {
	Id             int
	CurrencySymbol string
	DecimalDigits  int
	Slug           string
}

func CompanyHandler(next http.Handler) httprouter.Handle {
	return func(w http.ResponseWriter, r *http.Request, params httprouter.Params) {
		company := &Company{
			Slug: params[0].Value,
		}
		conn := getConn(r)
		err := conn.QueryRow(r.Context(), "select company_id, currency_symbol, decimal_digits from company join currency using (currency_code) where slug = $1", company.Slug).Scan(&company.Id, &company.CurrencySymbol, &company.DecimalDigits)
		if err != nil {
			http.NotFound(w, r)
			return
		}
		ctx := context.WithValue(r.Context(), ContextCompanyKey, company)
		r = r.WithContext(ctx)
		r2 := new(http.Request)
		*r2 = *r
		r2.URL = new(url.URL)
		*r2.URL = *r.URL
		r2.URL.Path = params[1].Value
		next.ServeHTTP(w, r2)
	}
}

func (c Company) MinCents() float64 {
	var r float64
	r = 1
	for i := 0; i < c.DecimalDigits; i++ {
		r /= 10.0
	}
	return r
}

func getCompany(r *http.Request) *Company {
	company := r.Context().Value(ContextCompanyKey)
	if company == nil {
		return nil
	}
	return company.(*Company)
}

type CurrencyOption struct {
	Code   string
	Symbol string
}

type CountryOption struct {
	Code string
	Name string
}

type Tax struct {
	Id   int
	Name string
	Rate int
}

type taxDetailsForm struct {
	*contactForm
	Currency *SelectField
}

func newTaxDetailsForm(ctx context.Context, conn *Conn, locale *Locale) *taxDetailsForm {
	return &taxDetailsForm{
		contactForm: newContactForm(ctx, conn, locale),
		Currency: &SelectField{
			Name:     "currency",
			Label:    pgettext("input", "Currency", locale),
			Options:  MustGetOptions(ctx, conn, "select currency_code, currency_symbol from currency order by currency_code"),
			Required: true,
			Selected: []string{"EUR"},
		},
	}
}

func (form *taxDetailsForm) Parse(r *http.Request) error {
	if err := form.contactForm.Parse(r); err != nil {
		return err
	}
	form.Currency.FillValue(r)
	return nil
}

func (form *taxDetailsForm) Validate(ctx context.Context, conn *Conn) bool {
	validator := newFormValidator()
	validator.CheckValidSelectOption(form.Currency, gettext("Selected currency is not valid.", form.locale))
	return form.contactForm.Validate(ctx, conn) && validator.AllOK()
}

func (form *taxDetailsForm) mustFillFromDatabase(ctx context.Context, conn *Conn, company *Company) *taxDetailsForm {
	err := conn.QueryRow(ctx, "select business_name, substr(vatin::text, 3), trade_name, phone, email, web, address, city, province, postal_code, country_code, currency_code from company where company_id = $1", company.Id).Scan(form.BusinessName, form.VATIN, form.TradeName, form.Phone, form.Email, form.Web, form.Address, form.City, form.Province, form.PostalCode, form.Country, form.Currency)
	if err != nil {
		panic(err)
	}
	return form
}

type TaxDetailsPage struct {
	DetailsForm *taxDetailsForm
	NewTaxForm  *taxForm
	Taxes       []*Tax
}

func GetCompanyTaxDetailsForm(w http.ResponseWriter, r *http.Request, _ httprouter.Params) {
	mustRenderTaxDetailsForm(w, r, newTaxDetailsFormFromDatabase(r))
}

func newTaxDetailsFormFromDatabase(r *http.Request) *taxDetailsForm {
	locale := getLocale(r)
	conn := getConn(r)
	form := newTaxDetailsForm(r.Context(), conn, locale)

	company := mustGetCompany(r)
	form.mustFillFromDatabase(r.Context(), conn, company)

	return form
}

func HandleCompanyTaxDetailsForm(w http.ResponseWriter, r *http.Request, _ httprouter.Params) {
	locale := getLocale(r)
	conn := getConn(r)
	form := newTaxDetailsForm(r.Context(), conn, locale)
	if err := form.Parse(r); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := verifyCsrfTokenValid(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	if ok := form.Validate(r.Context(), conn); !ok {
		w.WriteHeader(http.StatusUnprocessableEntity)
		mustRenderTaxDetailsForm(w, r, form)
		return
	}
	company := mustGetCompany(r)
	conn.MustExec(r.Context(), "update company set business_name = $1, vatin = ($11 || $2)::vatin, trade_name = $3, phone = parse_packed_phone_number($4, $11), email = $5, web = $6, address = $7, city = $8, province = $9, postal_code = $10, country_code = $11, currency_code = $12 where company_id = $13", form.BusinessName, form.VATIN, form.TradeName, form.Phone, form.Email, form.Web, form.Address, form.City, form.Province, form.PostalCode, form.Country, form.Currency, company.Id)
	http.Redirect(w, r, companyURI(company, "/tax-details"), http.StatusSeeOther)
	return
}

func mustRenderTaxDetailsForm(w http.ResponseWriter, r *http.Request, form *taxDetailsForm) {
	locale := getLocale(r)
	page := &TaxDetailsPage{
		DetailsForm: form,
		NewTaxForm:  newTaxForm(locale),
	}
	mustRenderTexDetailsPage(w, r, page)
}

func mustRenderTaxForm(w http.ResponseWriter, r *http.Request, form *taxForm) {
	page := &TaxDetailsPage{
		DetailsForm: newTaxDetailsFormFromDatabase(r),
		NewTaxForm:  form,
	}
	mustRenderTexDetailsPage(w, r, page)
}

func mustRenderTexDetailsPage(w http.ResponseWriter, r *http.Request, page *TaxDetailsPage) {
	conn := getConn(r)
	company := mustGetCompany(r)
	page.Taxes = mustGetTaxes(r.Context(), conn, company)
	mustRenderAppTemplate(w, r, "tax-details.gohtml", page)
}

func mustGetCompany(r *http.Request) *Company {
	company := getCompany(r)
	if company == nil {
		panic(errors.New("company: required but not found"))
	}
	return company
}

func mustGetTaxes(ctx context.Context, conn *Conn, company *Company) []*Tax {
	rows, err := conn.Query(ctx, "select tax_id, name, (rate * 100)::integer from tax where company_id = $1 order by rate, name", company.Id)
	if err != nil {
		panic(err)
	}
	defer rows.Close()

	var taxes []*Tax
	for rows.Next() {
		tax := &Tax{}
		err = rows.Scan(&tax.Id, &tax.Name, &tax.Rate)
		if err != nil {
			panic(err)
		}
		taxes = append(taxes, tax)
	}
	if rows.Err() != nil {
		panic(rows.Err())
	}

	return taxes
}

type taxForm struct {
	locale *Locale
	Name   *InputField
	Rate   *InputField
}

func newTaxForm(locale *Locale) *taxForm {
	return &taxForm{
		locale: locale,
		Name: &InputField{
			Name:     "tax_name",
			Label:    pgettext("input", "Tax name", locale),
			Type:     "text",
			Required: true,
		},
		Rate: &InputField{
			Name:     "tax_rate",
			Label:    pgettext("input", "Rate (%)", locale),
			Type:     "number",
			Required: true,
			Attributes: []template.HTMLAttr{
				"min=-99",
				"max=99",
			},
		},
	}
}

func (form *taxForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	form.Name.FillValue(r)
	form.Rate.FillValue(r)
	return nil
}

func (form *taxForm) Validate() bool {
	validator := newFormValidator()
	validator.CheckRequiredInput(form.Name, gettext("Tax name can not be empty.", form.locale))
	if validator.CheckRequiredInput(form.Rate, gettext("Tax rate can not be empty.", form.locale)) {
		validator.CheckValidInteger(form.Rate, -99, 99, gettext("Tax rate must be an integer between -99 and 99.", form.locale))
	}
	return validator.AllOK()
}

func HandleDeleteCompanyTax(w http.ResponseWriter, r *http.Request, params httprouter.Params) {
	taxId, err := strconv.Atoi(params[0].Value)
	if err != nil {
		http.NotFound(w, r)
		return
	}
	if err := verifyCsrfTokenValid(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	conn := getConn(r)
	conn.MustExec(r.Context(), "delete from tax where tax_id = $1", taxId)
	http.Redirect(w, r, companyURI(mustGetCompany(r), "/tax-details"), http.StatusSeeOther)
}

func HandleAddCompanyTax(w http.ResponseWriter, r *http.Request, _ httprouter.Params) {
	locale := getLocale(r)
	form := newTaxForm(locale)
	if err := form.Parse(r); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := verifyCsrfTokenValid(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	if !form.Validate() {
		w.WriteHeader(http.StatusUnprocessableEntity)
		mustRenderTaxForm(w, r, form)
		return
	}
	conn := getConn(r)
	company := mustGetCompany(r)
	conn.MustExec(r.Context(), "insert into tax (company_id, name, rate) values ($1, $2, $3 / 100::decimal)", company.Id, form.Name, form.Rate.Integer())
	http.Redirect(w, r, companyURI(company, "/tax-details"), http.StatusSeeOther)
}
